/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 *
 * Most of this file is code from the 3.0.1 viterbi package:
 * Copyright 1999 Phil Karn, KA9Q
 */

#include "codec.h"
#include "lib/conv.h"
#include "lib/cencoder_opt.h"

const char *progname = PARSEOPTS_PACKAGE;


/** -= gengetopt optstart =- *******************
   --func-name=parseopts --unamed-opts
*** -= gengetopt datastart =- ******************
   package "cencoder"

   version "1.15"

   purpose "Reads a stream of bits in network byte order from standard input.
   Convolutionally encode the bit stream and write it on standard output
   in bit mode, where 0x00 means '0' and 0xff means '1'."

   option "code" c "use the given known code (default=7,1/2)" string no
   option "list" l "list the known codes and exit" flag off
   option "debug" d "print some more error messages" flag off
*** -= gengetopt end =- ***********************/

/* 8-bit parity lookup table, generated by partab.c */
static int8_t Partab[] = {
  0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0, 1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1,
  1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1, 0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0,
  1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1, 0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0,
  0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0, 1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1,
  1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1, 0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0,
  0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0, 1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1,
  0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0, 1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1,
  1,0,0,1,0,1,1,0, 0,1,1,0,1,0,0,1, 0,1,1,0,1,0,0,1, 1,0,0,1,0,1,1,0
};


int
main (int argc, char *argv[])
{
  bool debug = FALSE;			/* printout */
  int K = 7;				/* constraint length */
  int k = 1;				/* code dimension */
  int n = 2;				/* codeword length */
  unsigned int *taps = NULL;		/* the n polynomials */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num != 0)
    {
      parseopts_print_help();
      fatal("no arguments required", NULL);
    }
  if (args_info.code_given)
    {
      if (sscanf(args_info.code_arg, "%u,%u/%u", &K, &k, &n) != 3)
  	fatal("wrong format for code", NULL);
    }
  if (args_info.list_given)
    {
      list_ccodes(stdout);		/* do a listing of the known codes */
      exit(0);				/* and nothing more */
    }
  if (args_info.debug_given)
    debug = TRUE;

  (void) init_stdio_bufs();


  /* Initialisation */
  for (struct ccode *cp = known_ccodes; cp->K > 0; cp++)
    if (cp->K == K && cp->k == k && cp->n == n)
      taps = cp->taps;
  if (taps == NULL)
    fatal("no known code for (%.9s)", args_info.code_arg);
  for (int i = 0; i < 256; i++)
    Partab[i] = 0 - Partab[i];


  /* Begin transmission */
  if (K == 7 && k == 1 && n == 2	/* (7,1/2) code */
      && taps[0] == 0x6d && taps[1] == 0x4f) /* the standard one */
    {
      /* This code copied from encode27.c by Phil Karn, which see. */

      int retval;
      unsigned int encstate = 0;	/* starting encoder state */
      const int poly1 = 0x6d;
      const int poly2 = 0x4f;

      while ((retval = getchar()) != EOF)
	for (int i = 7; i >= 0; i--)
	  {
	    encstate = (encstate << 1) | ((retval >> 7) & 1);
	    retval <<= 1;
	    if (putchar(Partab[encstate & poly1]) == EOF)
	      pfatal("error writing on output");
	    if (putchar(Partab[encstate & poly2]) == EOF)
	      pfatal("error writing on output");
	  }

      /* Generate tail */
      for (int i = 5; i >= 0; i--)
	{
	  encstate = encstate << 1;
	  if (putchar(Partab[encstate & poly1]) == EOF)
	    pfatal("error writing on output");
	  if (putchar(Partab[encstate & poly2]) == EOF)
	    pfatal("error writing on output");
	}
    }
  else
    fatal("code not implemented", NULL);

  if (debug && feof(stdin))
    error("end of file reading from input", NULL);
  if (ferror(stdin))
    pfatal("error reading from input");

  return 0;
}


/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
