/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "codec.h"
#include "lib/channel_opt.h"

const char *progname = PARSEOPTS_PACKAGE;

static double clearsky(double sample);	/* no noise */
static double awgn(double sample);	/* additive white gaussian noise */
static double mtrace(double sample);	/* noise trace file */


/** -= gengetopt optstart =- *******************
   --func-name=parseopts --unamed-opts
*** -= gengetopt datastart =- ******************
   package "channel"

   version "1.31"

   purpose "for each byte read from standard input, write a byte on
   standard output.  A 0x00 byte on input means the number -1,
   while a 0xff byte means the number 1.
   The input number is modified by noise, then quantised and compressed.
   If no quantisation bits are specified, use hard quantisation.  Else,
   multiply the result by comp and quantise it in uniform quantisation
   regions.  Output numbers are in binary offset format: 0 is invalid,
   1-127 is a 0 symbol, 128 is an erasure, 129-255 is a 1 symbol.
   Noise types are:
   	clearsky: no noise (this is the default)
   	awgn,<ebno>: additive white Gaussian noise, channel Eb/No in dB
   	mtrace,<file>: multiplicative noise, file contains a sequence of
   		bytes in binary offset format: 0xff means 1, 0x01 means -1."

   option "qbit" q "quantisation bits in [1-8] (default=1)" int no
   option "comp" c "compression factor (default=1)" float no
   option "noise" n "noise type (default=clearsky)" string no
   option "seed"  s "a 32-bit integer used as a seed (default=4357)" long no
   option "debug" d "print generator polynomial on stderr" flag off
*** -= gengetopt end =- ***********************/

static union
{
  struct {
    double sd;				/* sample standard deviation */
  } awgn;				/* data for additive white Gaussian noise  */
  struct {
    FILE *stream;			/* the trace file */
  } mtrace;
} noiseargs;

static bool debug = FALSE;

int
main (int argc, char *argv[])
{
  int qbit = 1;				/* default is hard quantisation */
  double comp = 1;			/* default compression is none */
  double (*noise)(double) = clearsky;	/* default is no noise added */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num != 0)
    {
      parseopts_print_help();
      fatal("no arguments required", NULL);
    }
  if (args_info.seed_given)
    seedMT((uint32)args_info.seed_arg);	/* seed the generator */
  if (args_info.qbit_given)		/* quantisation bits option */
    {
      qbit = args_info.qbit_arg;
      if (qbit < 1 || qbit > 8)
	fatal("bad argument to qbit option", NULL);
    }
  if (args_info.comp_given)		/* compression factor option */
    {
      comp = args_info.comp_arg;
      if (comp < 0.01 || comp > 100)
	fatal("bad argument to comp option", NULL);
      if (comp > 1)
	error("warning: compression factor greater than 1", NULL);
    }
  if (args_info.noise_given)
    {
      if (noise != clearsky)
	fatal("more than one noise type selected", NULL);
#     define noiseis(string) !strncmp(string, args_info.noise_arg, sizeof(string)-1)

      if (noiseis("clearsky\0"))
	noise = clearsky;		/* do not add any noise */
      else if (noiseis("awgn,"))
	{
	  noise = awgn;			/* use additive white Gaussian noise */
	  float ebno;

	  if (sscanf(args_info.noise_arg, "awgn,%g", &ebno) != 1
	      || ebno < -20
	      || ebno > 20)
	    fatal("bad argument to awgn option -- %s", args_info.noise_arg);
	  noiseargs.awgn.sd = sqrt(0.5 * pow(10, -ebno/10));
	}
      else if (noiseis("mtrace,"))
	{
	  noise = mtrace;		/* use multiplicative noise from trace */
	  char *tracename;		/* name of trace file */

	  tracename = args_info.noise_arg + strlen("mtrace,");
	  if ((noiseargs.mtrace.stream = fopen(tracename, "readonly")) == NULL)
	    pfatal("while opening mtrace file");
	}
      else
	fatal("unknown noise type selected - %s", args_info.noise_arg);
    }
  if (args_info.debug_given)
    debug = TRUE;

  (void) init_stdio_bufs();

  int retval;
  while ((retval = getchar()) != EOF)
    {
      byte in = retval;			/* input sample as '0' or '1' */
      double sample;
      const bool accept_bitmode_only = TRUE;
      switch (in)
	{
	case 0x00:
	  sample = -1;
	  break;
	case 0x01:
	  if (accept_bitmode_only)
	    goto not_bit_mode_error;
	  sample = -1;		       /* promote to -1 rather than -127/128 */
	  break;
	case 0xff:
	  sample = 1;			/* promote to 1 rather than 127/128 */
	  break;
	not_bit_mode_error:
	default:
	  if (accept_bitmode_only)
	    fatal("input stream is not in bit mode", NULL);
	  sample = (double)in / OFFSET - 1;
	}
      sample = noise(sample);

      sample *= comp;			/* compress */
      int oval;				/* the output number */
      if (sample == 0)			/* unlikely case */
	oval = OFFSET;			/* an erasure */
      else
	{
	  oval = 1 + fabs(sample) * (1<<(qbit-1)); /* quantize */
	  oval *= (OFFSET>>(qbit-1));	/* normalize */
	  if (oval > OFFSET-1)
	    oval = OFFSET-1;		/* truncate */
	  if (sample > 0)		/* a `1' soft symbol */
	    oval += OFFSET;		/* put in [129-255] */
	  else				/* a `0' soft symbol */
	    oval = OFFSET - oval;	/* put in [127-1] */
	}
      assert(oval > 0 && oval < 256);

      byte out = oval;			/* output sample as binary-offset */
      if (putchar(out) == EOF)
	pfatal("error writing on output");
    }
  if (debug && feof(stdin))
    error("end of file reading from input", NULL);
  if (ferror(stdin))
    pfatal("error reading from input");

  return 0;
}


/*
 * No noise.
 */
static double
clearsky(double sample)
{
  return sample;
}


/*
 * Additive white Gaussian noise.
 */
static double
awgn(double sample)
{
  return sample + noiseargs.awgn.sd * normal();
}


/*
 * Multiplicative error from trace file.
 */
static double
mtrace(double sample)
{
  int bonoise = getc(noiseargs.mtrace.stream);	/* binary offset noise */
  if (bonoise == EOF)
    {
      if (ferror(noiseargs.mtrace.stream))
	pfatal("while reading mtrace file"); /* does not return */
      if (debug)
	error("end of file reading from mtrace file", NULL);
      exit(0);
    }

  double noise = (bonoise - 128) / 127;	/* transform to 2's complement */

  return sample * noise;
}


/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
