#! /bin/bash
#  A suite of programs for simulating channel codings
#  Copyright (C) 2002,2003,2005 Francesco Potort
#  See file COPYING for details

if [ $# -lt 3 ]; then
    cat <<EOF
usage: $0 [options] <range|file> <outfile> <errfile>
	range is <min>:<step>:<max> in dB or multiplicative error trace
	outfile and errfile can be - for stdin and stdout
	see the source for the available options
EOF
    exit 2
fi
LC_NUMERIC=C				# use C style for numbers

errlimit=10				# default error bursts to generate
ccode=7,1/2				# default convolutional code
conflev=0.9				# default confidence level
decmem=80				# default decoding memory depth
ppat=none				# default puncturing pattern
qlevopt='-q8 -c.5'			# default is soft decoding
rscode=none				# default Reed-Solomon code

while getopts "e:c:l:m:p:q:rs:" opt; do
    case $opt in
    "e")
	    errlimit="$OPTARG" ;;
    "c")
	    ccode="$OPTARG" ;;
    "l")
	    conflev="$OPTARG" ;;
    "m")
	    decmem="$OPTARG";;
    "p")
	    ppat="$OPTARG" ;;
    "q")
	    qlevopt="$OPTARG" ;;
    "r")
	    rscode=yes ;;
    "s")
	    seed="-s $OPTARG";;
    *)
	    exit 1 ;;
    esac
done
shift $((OPTIND-1))			# shift options away
range=$1				# dB range as start:step:limit
out=$2					# output file
errors=$3				# errors file

mtrace="$range";			# multiplicative error trace
if [ -r "$mtrace" ]; then		# yes, we have channel errors trace
    noise="mtrace,$mtrace"		# option for the channel program
    start=0; step=1; limit=0;		# one run only
    errlimitopt=			# no error limit
else
    start=$(echo $range | cut -d: -f1)	# min dB
    step=$(echo $range | cut -d: -f2)	# dB step
    limit=$(echo $range | cut -d: -f3)	# max dB
    mtrace=				# no multiplicative error trace
    errlimitopt="-e $errlimit"
fi


if [ "$ccode" = none ]; then		# no convolutional encoding
    cencoder="rangen -b"
    vdecoder=harddec
    rate=1
    codename="1/1"
    group=
    burstlen="          "
else
    cencoder="cencoder -c $ccode"	# convolutional encoder command
    vdecoder="vdecoder -c $ccode -m $decmem" # viterbi decoder command
    cl=$(echo $ccode | cut -d, -f1)	# constraint lenght
    k=$(echo $ccode | tr ,/ \\t | cut -f2) # input code block size
    n=$(echo $ccode | cut -d/ -f2)	# output code block size
    codename="($ccode)"			# pretty print of the code name
    rate=$(echo "$k / $n"|bc -l)	# convolutional code rate
    bcl=$((k*cl))		     # constraint length (in input bits)
    group="-c $bcl"			# group option for printerr
    burstlen="errs/burst"
fi

if [ "$ppat" = none ]; then		# do not use puncturing
    enpuncture=cat
    depuncture=cat
else
    pcode=$(echo $ppat | cut -d= -f1)	# convolutional code
    if [ "$pcode" != "$ccode" ]; then exit 1; fi
    pkpn=$(echo $ppat | cut -d= -f2)
    pk=$(echo $pkpn | cut -d/ -f1)	# rate numerator
    pn=$(echo $pkpn | cut -d/ -f2)	# rate denominator
    enpuncture="enpuncture -c $ppat"	# puncturing command
    depuncture="depuncture -c $ppat"	# depuncturing command
    codename="($ppat)"
    rate=$(echo "$pk / $pn"|bc -l)	# punctured code rate
    group="-c $bcl"			# group option for printerr
    burstlen="errs/burst"
fi

if [ "$rscode" = none ]; then
    rsencoder=cat
    rsdecoder=cat
else
    rsn=255				# RS block size
    rsk=223				# RS data symbols
    rsencoder="rsencoder"		# RS encoding command
    rsdecoder="rsdecoder"		# RS decoding command
    codename="($rsn,$rsk)+$codename"
    rate=$(echo "$rate * $rsk/$rsn"|bc -l) # add Reed-Solomon code rate
    group="-b $((rsk*8))"		# group option for printerr
    burstlen="errs/block"
fi


# fromdB: e(x/10*l(10))
# todB:   10*l(x)/l(10)
gain=$(echo "-10*l($rate)/l(10)"|bc -l)	# code gain in dB
    
if [ -n "$out" -a "$out" != - ]; then
    : rewind out file > $out
else
    out=/dev/stdout
fi
exec 3>$out; out=/dev/fd/3
if [ "$errors" ]; then
    : rewind errors file > $errors
fi

epoch=$(date +%s)			# get starting time
bitcount=.TMPcodec.bitcount.$$		# contains total number of bits
trap "rm -f $bitcount" EXIT

printf "BER test for the $codename code with %.0f%% confidence: \n" \
       $(echo "$conflev * 100"|bc -l) >>$out
printf " Eb/No Ec/No $burstlen        BER \n" >>$out

bits=0;
ebno=$start;
while [ $(echo "$ebno <= $limit"|bc -l) = 1 ]; do

    if [ "$mtrace" ]; then		# we have a multiplicative error trace
	ecno=0
    else
	ecno=$(echo "$ebno-$gain"|bc -l) # channel Eb/No
	noise="awgn,$ecno"
    fi

    printf "%6.2f%6.2f" $ebno $ecno >>$out

    test "$debug" && echo "
    rangen $seed |
        $rsencoder |
	$cencoder |
	$enpuncture |
	channel $qlevopt --noise=$noise |
	$depuncture |
	$vdecoder |
	$rsdecoder |
	printerr $group -l $conflev $errlimitopt <(rangen $seed) - $out -"

    rangen $seed |
        $rsencoder |
	$cencoder |
	$enpuncture |
	channel $qlevopt --noise=$noise |
	$depuncture |
	$vdecoder |
	$rsdecoder |
	printerr $group -l $conflev $errlimitopt <(rangen $seed) - $out - |
	tee -a $errors |
	tail -1 | cut -f 2 >$bitcount

    bits=$(echo "$bits+$(cat $bitcount)"|bc -l)

    ebno=$(echo "$ebno + $step"|bc -l)
done

# Print elapsed time and estimated speed
{
  rt=$(($(date +%s)-epoch))
  echo -n Total $bits generated symbols, runtime
  printf " %02d:%02d hours, sym/hour %.2e.\n" \
	 $((rt/3600)) $(((rt/60)%60)) $(echo "$bits/$rt*3600"|bc -l)
} >>$out
