/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "codec.h"
#include "lib/conv.h"
#include "lib/enpuncture_opt.h"

const char *progname = PARSEOPTS_PACKAGE;


/** -= gengetopt optstart =- *******************
   --func-name=parseopts --unamed-opts
*** -= gengetopt datastart =- ******************
   package "enpuncture"

   version "1.5"

   purpose "Reads a stream of bytes in bit mode (where 0x00 means '0'
   and 0xff means '1') from standard input.  Apply a puncturing pattern
   to the stream and write it on standard output in bit mode."

   option "pattern" c "use the given known pattern (default=7,1/2=3/4)" string no
   option "list" l "list the known erasure patterns and exit" flag off
   option "debug" d "print some more error messages" flag off
*** -= gengetopt end =- ***********************/


int
main (int argc, char *argv[])
{
  bool debug = FALSE;			/* printout */
  char *pattern = "7,1/2=3/4";		/* default puncturing pattern */
  int p = 0;				/* period (bits) */
  bool *erasures = NULL;	      /* the erasures: 1 is pass, 0 is erase */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num != 0)
    {
      parseopts_print_help();
      fatal("no arguments required", NULL);
    }
  if (args_info.pattern_given)
    {
      pattern = args_info.pattern_arg;
      unsigned int d;
      if (sscanf(pattern, "%u,1/%u=%u/%u", &d, &d, &d, &d) != 4)
  	fatal("wrong format for pattern", NULL);
    }
  if (args_info.list_given)
    {
      list_ppats(stdout);		/* do a listing of the known codes */
      exit(0);				/* and nothing more */
    }
  if (args_info.debug_given)
    debug = TRUE;

  (void) init_stdio_bufs();


  /* Initialisation */
  for (struct ppat *pp = known_ppats; pp->K > 0; pp++)
    if (!strcmp(pattern, pp->name))
      {
	erasures = make_erasure_list(pp);
	p = pp->P * pp->n;
      }
  if (p == 0)
    fatal("no known pattern for (%.15s)", args_info.pattern_arg);


  /* Begin transmission */
  int retval;
  int i = 0;
  while ((retval = getchar()) != EOF)
    {
      switch (retval)
	{
	case 0x00: case 0xff: break;
	default:
	  fatal("input stream is not in bit mode", NULL);
	  /* NOTREACHED */
	}
      if (erasures[i])
	if (putchar(retval) == EOF)
	  pfatal("error writing on output");
      if (++i == p)
	i = 0;
    }
  if (debug && feof(stdin))
    error("end of file reading from input", NULL);
  if (ferror(stdin))
    pfatal("error reading from input");

  return 0;
}


/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
