/*
 * A program for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "../codec.h"

#include <float.h>

/* Inverse error function.  Copied from Octave, GPL, by Kurt Hornik 1996 */
static inline double
erfinv(double x)
{
  if (x < -1 || x > 1)
    return NAN;
  if (x == -1)
    return -INFINITY;
  if (x == 1)
    return INFINITY;
  int iterations = 0;
  double z_old;
  double z_new = copysign(sqrt(-log(1-fabs(x))), x);
  while (fabs(erf(z_new)-x) > DBL_EPSILON*fabs(x)
	 && iterations++ < 100)		/* max iterations */
    {
      z_old = z_new;
      z_new = z_old - (erf(z_old)-x) * exp(z_old*z_old) / M_2_SQRTPI;
    }
  return z_new;
}

/*
 * Given conflev (a quantile), returns the sigma such that
 *   conflev = Integral[-sigma, sigma, N(0,1)],
 * where N(0,1) is the normal distribution with null mean and unitary
 * variance.
 */
double
invQ(double conflev)
{
  return M_SQRT2 * erfinv(conflev);
}


/*
 * Use the Box-Muller algorithm to produce a Gaussian random variable
 * with zero mean and unit variance.
 */
double
normal(void)
{
  static bool sparegood = FALSE;
  static double spareval;

  if (sparegood)			/* no need to compute now */
    {
      sparegood = FALSE;
      return spareval;
    }

  double r2, vx, vy;

  do					/* do compute one now */
    {
      const double k = 2 / (double)MTMAXRAND;
      vx = k * randomMT() - 1;
      vy = k * randomMT() - 1;
      r2 = vx * vx + vy * vy;
    }
  while (r2 >= 1 || r2 == 0);		/* until we are in the unit circle */

  /* Box-Muller transformation. */
  double m = sqrt(-2 * log(r2) / r2);
  spareval = vx * m;			/* spare value */
  sparegood = TRUE;			/* we have one */

  return vy * m;			/* return the other */
}


bool
coin()
{
  static int c = 1;
  static uint32 r;

  if (--c == 0)
    {
      c = MTSIZE;
      r = ~randomMT();
    }
  else
    r <<= 1;
  return !(r & 0x80000000u);
}



/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
