/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "codec.h"
#include "lib/printerr_opt.h"

const char *progname = PARSEOPTS_PACKAGE;

static void compare_bits(byte *p1, byte *p2, size_t len);
static void print_results(void);

/**** -= gengetopt optstart =- ********************************
   --func-name=parseopts --unamed-opts
***** -= gengetopt datastart =- *******************************
   package "printerr"

   version "1.40"

   purpose "compares two streams of data from the two files whose names are given
   as first and second arguments, where `-' means standard input.
   The third and fourth arguments are names of files (`-' means standard
   output) opened in write append mode.  Final results go to the third
   arg file, the list of errors go to the fourth arg file.
     If the streams are composed of the bytes 0x00 and 0xff, assumes
   one bit per input byte.
     If an error limit is given, stop after that number of errors.
     After the given clen unerrored bits assumes that the error burst
   ended, and if errlimit is given, stop after that number of bursts.
     Alternatively, if errlimit and blen are given, stop after that
   number of errored blocks.
     Without any of clen or blen, only consider the number of errors.
     The final results are the average number of errors per burst or
   block, the BER and the confidence interval.  If clen/blen is given,
   the confidence interval is computed on the number of bursts/blocks.
     If no fourth arg is specified, the list of errors is not written,
   else it is the sequence number of differing bits of the two streams,
   preceded by the burst/block sequence number if clen/blen is given."

   option "blen" b "block size in bits, used for stopping" int no
   option "clen" c "constraint length, used for separating error bursts" int no
   option "errlimit" e "stop after this number of errors" int no
   option "clevel" l "confidence level (default is 90%%)" float no
   option "debug" d "print some more error messages" flag off
***** -= gengetopt end =- ************************************/

static bool debug = FALSE;		/* printout */
static bool bitmode = TRUE;		/* bit (as opposed to byte) mode */

static float conflev = 0.9;		/* confidence level = 90% */
static unsigned int clen = 0;		/* constraint length */
static unsigned int blen = 0;		/* block size */
static unsigned int errlimit = 0;	/* limit number of errors */

static FILE *infile[2];			/* the streams */
static char *inbuf[2];			/* the stream buffers */
static FILE *outfile;			/* results */
static FILE *errfile;			/* errors count */

static bigcnt N;			/* bits read */
static bigcnt bursts;			/* number of error burst */
static bigcnt blocks;			/* number of errored blocks */
static bigcnt errors;			/* number of errored bits */



static void
print_results(void)
{
  /* Print error rate results */
  if (errors > 0)
    {
      double ber = (double) errors / N;
      double ss = invQ(conflev);	/* +/-sigma for conflev */
      bigcnt groups = bursts+blocks;	/* one or both are zero */
      assert(bursts*blocks == 0);
      if (groups > 0)
	fprintf(outfile, " %10.1f %10.1e +/- %2d%%\n",
		(double)errors/groups,
		ber, (int)floor(0.5+ss*100/sqrt(groups)));
      else
	fprintf(outfile, "            %10.1e +/- %2d%%\n",
		ber, (int)floor(0.5+ss*100/sqrt(errors)));
    }
  else
    {
      double point = -log(1-conflev);
      fprintf(outfile, " %10s < %8.1e\n", "", point/N);
    }
  fflush(outfile);

  exit(0);
  /* NOTREACHED */
}



static void
compare_bits (byte *p1, byte *p2,	/* arrays of len bytes */
	      size_t len)		/* size of data to compare */
{
  size_t nbit;				/* bit counter */
  size_t inc = (bitmode) ? 1 : 8;	/* bit counter increment */
  size_t end = (bitmode) ? len : len*8;	/* bit counter limit */

  for (nbit = 0; nbit < end; nbit += inc, p1++, p2++)
    {
      size_t b = nbit;			/* number of errored bit */

      while (*p1 != *p2)		/* differing bytes */
	{
	  if (bitmode)			/* only one errored bit */
	    {
	      b += 1;			/* bit numbers are 1-based */
	      *p1 = *p2;		/* reset the difference */
	    }
	  else
	    {
	      int noerror;		/* first bits are the same */
	      do
		{
		  noerror = ((*p1 & 0x80) == (*p2 & 0x80));
		  b += 1;		/* inc number of errored bit */
		  *p1 <<= 1, *p2 <<= 1;	/* go to next bit */
		}
	      while (noerror);		/* until first bits differ */
	    }

	  if (clen > 0)
	    {
	      static bigcnt bitno;	/* last errored bit */
	      if (bursts == 0		/* no previous burst of errors */
		  || N+b-bitno > clen)	/* constraint length exceeded */
		{
		  if (errlimit > 0	/* burst limit requested */
		      && bursts == errlimit) /* burst limit exceeded */
		    {
		      N += b-1;		/* update number of bits read */
		      print_results();
		      /* NOTREACHED */
		    }
		  bursts += 1;		/* a new burst (1-based) */
		}
	      bitno = N+b;		/* save last errored bit number */
	      if (errfile)
		fprintf(errfile, "%llu\t", bursts); /* print number of bursts */
	    }
	  else if (blen > 0)
	    {
	      static bigcnt blockno;	/* last errored block number */
	      bigcnt bno = (N+b)/blen+1; /* this block number */
	      if (bno > blockno)	/* a different block */
		{
		  if (errlimit > 0	/* burst limit requested */
		      && blocks == errlimit) /* burst limit exceeded */
		    {
		      N += b-1;		/* update number of bits read */
		      print_results();
		      /* NOTREACHED */
		    }
		  blockno = bno;	/* save last errored block number */
		  blocks += 1;		/* a new block (1-based) */
		}
	      if (errfile)
		fprintf(errfile, "%llu\t", blocks); /* print number of blocks */
	    }

	  if (errfile)
	    fprintf(errfile, "%llu\n", N+b); /* print errored bit number */
	  errors += 1;			/* increment the global counter */

	  if (errors == errlimit && clen == 0 && blen == 0)
	    {
	      N += b;			/* update number of bits read */
	      print_results();
	      /* NOTREACHED */
	    }
	}
    }
  N += nbit;				/* update number of bits read */

  if (errlimit > 0			/* error/burst limit requested */
      && ((clen > 0			/* work with bursts */
	   && bursts >= errlimit)	/* burst limit reached or exceeded */
	  || (blen > 0			/* work with blocks */
	      && blocks >= errlimit)	/* block limit reached or exceeded */
	  || (clen == 0 && blen == 0	/* work with errors */
	      && errors >= errlimit)))  /* error limit reached or exceeded */
    {
      print_results();
      /* NOTREACHED */
    }
  if (errfile)
    fflush(errfile);
}


int
main (int argc, char *argv[])
{
  char *filename;			/* name of currently used file */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num < 3)
    {
      parseopts_print_help();
      fatal("three arguments required", NULL);
    }
  if (args_info.inputs_num > 4)
    {
      parseopts_print_help();
      fatal("more arguments than required", NULL);
    }
  if (args_info.blen_given)
    {
      if (args_info.blen_arg <= 0)
	fatal("block length must be positive", NULL);
      blen = args_info.blen_arg;
    }
  if (args_info.clen_given)
    {
      if (args_info.blen_given)
	fatal("blen and clen are alternative", NULL);
      if (args_info.clen_arg <= 0)
	fatal("constraint length must be positive", NULL);
      clen = args_info.clen_arg;
    }
  if (args_info.clevel_given)
    {
      if (args_info.clevel_arg <= 0
	  || args_info.clevel_arg >= 1)
	fatal ("confidence level must be in ]0,1[", NULL);
      conflev = args_info.clevel_arg;
    }
  if (args_info.errlimit_given)
    {
      if (args_info.errlimit_arg <= 0)
	fatal ("max number of %s must be positive",
	       (clen > 0) ? "bursts" : ((blen > 0) ? "blocks" : "errors"));
      errlimit = args_info.errlimit_arg;
    }
  if (args_info.debug_given)
    debug = TRUE;


  /* Open the output files */
  filename = args_info.inputs[2];
  if (filename[0] == '-' && filename[1] == '\0')
    outfile = stdout;
  else
    outfile = fopen(filename, "a");
  if (outfile == NULL)
    goto open_error;

  if (args_info.inputs_num == 4)
    /* error count requested */
    {
      filename = args_info.inputs[3];
      if (filename[0] == '-' && filename[1] == '\0')
	errfile = stdout;
      else
	errfile = fopen(filename, "a");
      if (errfile == NULL)
	goto open_error;
    }

  const size_t len = BUFSIZ;
  /* Lenght of chunk of data in bytes MUST be a power of 2! */
  for (size_t l = len; (l & 1) == 0; l >>= 1)
    assert((l&2)==0 || l==2);

  /* Open the input streams and allocate the input buffers */
# define for_both_streams(s) for (int s=0; s<2; s++)
  for_both_streams(s)
    {
      /* Open the file */
      filename = args_info.inputs[s];
      if (filename[0] == '-' && filename[1] == '\0')
	infile[s] = stdin;
      else
	infile[s] = fopen(filename, "r");
      if (infile[s] == NULL)
	{
	  char err[200];
	open_error:
	  sprintf(err, "error opening \"%.160s\"", filename);
	  pfatal(err);
	}

      /* Allocate the input buffer, on a page boundary */
      inbuf[s] = xnew(2*len-1, char);
      size_t misalign = (size_t)inbuf[s] & (len-1);
      if (misalign != 0)		/* if not aligned on page boundary */
	inbuf[s] += len-misalign;	/* realign it */
      setbuffer(infile[s], inbuf[s], len); /* set buffered input buffer */
    }


  /* Identify the type of data: bit or byte mode */
  byte in[2][len];			/* data read in */
  size_t nread[2];			/* amount of data read */
  for_both_streams(s)
    {
      byte *p = in[s];
      nread[s] = fread(p, 1, len, infile[s]);
      if (nread[s] != len && ferror(infile[s]))
	{
	  filename = args_info.inputs[s];
	  goto read_error;
	}
      for (int b = 0; b < nread[s]; b++)
	switch (p[b])
	  {
	  case 0x00: case 0xff: break;
	  default: bitmode = FALSE;
	  }
    }


  /* Read and compare the input data */
  int batch = (bitmode) ? len : len*8;	/* bits read each loop */
  while (nread[0] == len && nread[1] == len) /* common case */
    {
      if (memcmp(in[0], in[1], len))	/* uncommon case */
	compare_bits(in[0], in[1], len); /* compare and print results */
      else
	N += batch;			/* bits read this time */
      for_both_streams(s)
	nread[s] = fread(in[s], 1, len, infile[s]);
    }

  /* EOF or read error */
  for_both_streams(s)
    if (ferror(infile[s]))
      {
	char err[200];
	fflush(outfile);
	if (errfile) fflush(errfile);
	filename = args_info.inputs[s];
      read_error:
	sprintf(err, "error reading from \"%.160s\"", filename);
	pfatal(err);
      }

  /* If burst limit specified, normally does not exit from here */
  if (debug || errlimit > 0)
    error("%s input closed before reaching burst limit",
	  (feof(infile[0]) && feof(infile[1])) ? "both"
	  : (feof(infile[0])) ? "first"
	  : (feof(infile[1])) ? "second"
	  : "unknown");

  /* EOF on one or both streams */
  int min = nread[0];
  if (min > nread[1])
    min = nread[1];
  for_both_streams(s)
    {
      byte *p = in[s];
      for (int i = min; i < len; i++)
	p[i] = 0;			/* clear unread bytes */
    }
  if (memcmp(in[0], in[1], min))	/* uncommon case */
    compare_bits(in[0], in[1], min);	/* compare and print results */
  else
    N += (bitmode) ? min : min*8;

  print_results();
  /* NOTREACHED */

  return 0;
}

/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
