/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "codec.h"
#include "lib/rangen_opt.h"

const char *progname = PARSEOPTS_PACKAGE;

static size_t fill_buffer_text(byte* outbuf, int bits);
static size_t fill_buffer_bin(byte* outbuf, int bits);


/** -= gengetopt optstart =- *******************
   --func-name=parseopts --unamed-opts
*** -= gengetopt datastart =- ******************
   package "rangen"

   version "1.17"

   purpose "generates a random stream of bits using the Mersenne Twister.
   The number of bits generated is given as the count option,
   possibly followed by k, M, G, T, meaning 1e3, 1e6, 1e9, 1e12.
   No count option means forever.
   When generating binary data (the default) the first generated
   bit is the most significant bit of the first byte."

   option "bitmode" b "one char (0x00 or 0xff) per generated bit" flag off
   option "count" c "the number of bit to generate" string no
   option "seed"  s "a 32-bit integer used as a seed" long no
*** -= gengetopt end =- ***********************/

int
main (int argc, char *argv[])
{
  const size_t outbuf_sz = BUFSIZ;
  bigcnt N = 0;				/* number of bits to generate */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);
  if (args_info.inputs_num != 0)
    goto wrong_args;

  if (args_info.seed_given)
    seedMT((uint32)args_info.seed_arg);	/* seed the generator */

  if (!args_info.count_given)		/* no count option */
    N = ULLONG_MAX;			/* should be infinite */
  else					/* count option given */
    {
      char scale;			/* multiplier */
      int n = sscanf(args_info.count_arg, "%llu%c", &N, &scale);
      if (n < 1)
	goto wrong_args;
      if (n == 2)			/* multiplier given */
	switch (scale)
	  {
	  case 'T': if (N > ULLONG_MAX/1000) goto wrong_args; N *= 1000;
	  case 'G': if (N > ULLONG_MAX/1000) goto wrong_args; N *= 1000;
	  case 'M': if (N > ULLONG_MAX/1000) goto wrong_args; N *= 1000;
	  case 'k': if (N > ULLONG_MAX/1000) goto wrong_args; N *= 1000;
	    break;
	  default:			/* illegal multiplier */
	  wrong_args:
	    parseopts_print_help();
	    fatal("wrong argument", NULL);
	  }
    }

  size_t pagesize = getpagesize();
  assert(outbuf_sz % pagesize == 0);

  byte *outbuf = xnew((outbuf_sz+pagesize), byte);
  size_t misalign = (size_t)outbuf & (pagesize-1);
  if (misalign != 0)		   /* if not aligned on page boundary */
    outbuf += pagesize-misalign;	/* realign it */

  /* Compute the number of output buffers to fill and number of bits
     in the last buffer */
  int bitsperbuf = outbuf_sz;		/* how many bit in the buffer */
  if (!args_info.bitmode_flag)		/* 8 bits per byte in the buffer */
    bitsperbuf *= 8;

  /* Loop over buffers to fill */
  for (bigcnt buf = N/bitsperbuf + 1; buf > 0; buf--)
    {
      /* Fill the output buffer */
      int bits = bitsperbuf;		/* bits in this buffer */
      if (buf == 1)			/* last buffer */
	bits = N % bitsperbuf;		/* may be 0 */
      size_t outbytes;			/* bytes to write on stdout */

      if (args_info.bitmode_flag)
	outbytes = fill_buffer_text(outbuf, bits); /* 1 bit per output byte */
      else
	outbytes = fill_buffer_bin(outbuf, bits); /* 8 bits per output byte */

      /* Write the result on stdout */
      ssize_t retval;
      for (byte *ob = outbuf;
	   outbytes > 0;
	   outbytes -= retval, ob += retval)
	{
	  do
	    retval = write(STDOUT_FILENO, ob, outbytes);
	  while (retval < 0 && errno == EINTR);
	  if (retval < 0)
	    pfatal("write failed");
	}
    }

  return 0;
}


static size_t
fill_buffer_text(byte* outbuf, int bits)
{
  const byte table [16][4] = {
    {000,000,000,000}, {000,000,000,255}, {000,000,255,000}, {000,000,255,255},
    {000,255,000,000}, {000,255,000,255}, {000,255,255,000}, {000,255,255,255},
    {255,000,000,000}, {255,000,000,255}, {255,000,255,000}, {255,000,255,255},
    {255,255,000,000}, {255,255,000,255}, {255,255,255,000}, {255,255,255,255},
  };

  uint32 *b4p = (uint32 *)outbuf;   /* word pointer into buf */

  for (int i = 0; i < (bits+MTSIZE-1) / MTSIZE; i++)
    {
      uint32 r = randomMT();
      b4p[7] = ((uint32 *)table)[r & 0x0f];
      b4p[6] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[5] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[4] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[3] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[2] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[1] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p[0] = ((uint32 *)table)[(r >>= 4) & 0x0f];
      b4p += 8;
    }

  return bits;
}


static size_t
fill_buffer_bin(byte* outbuf, int bits)
{
  byte *b1p = outbuf;			/* byte pointer into buf */
  uint32 *b4p = (uint32 *)outbuf;	/* word pointer into buf */

  if (__BYTE_ORDER == __BIG_ENDIAN)
    for (int i = 0; i < bits / MTSIZE; i++)
      *b4p++ = randomMT();

  else				/* little endian */
    for (int i = 0; i < bits / MTSIZE; i++)
      {
	uint32 r = randomMT();
	b1p[3] = r & 0xff;
	b1p[2] = (r >>= 8) & 0xff;
	b1p[1] = (r >>= 8) & 0xff;
	b1p[0] = (r >>= 8) & 0xff;
	b1p += 4;
      }

  int residual = bits % MTSIZE;
  if (residual > 0)
    /* zero down the unrequested (least significant) bits */
    {
      uint32 r = randomMT() & (~(uint32)0 << (MTSIZE-residual));

      if (__BYTE_ORDER == __BIG_ENDIAN)
	*(uint32 *)b4p = r;
      else 			/* little endian */
	{
	  b1p[3] = r & 0xff;
	  b1p[2] = (r >>= 8) & 0xff;
	  b1p[1] = (r >>= 8) & 0xff;
	  b1p[0] = (r >>= 8) & 0xff;
	}
    }

  return (bits+7)/8;
}


/*
 * Local Variables:
 * comment-column: 40
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte")
 * End:
 */
