/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 */

#include "codec.h"
#include "rs.h"
#include "lib/rsdecoder_opt.h"

const char *progname = PARSEOPTS_PACKAGE;

/**** -= gengetopt optstart =- ********************************
   --func-name=parseopts --unamed-opts
***** -= gengetopt datastart =- *******************************
   package "rsdecoder"

   version "1.6"


   purpose "Read from standard input a bit stream containing Reed-Solomon
   redundancy information, if necessary and possible correct it,
   strip the redundancy information and write the resulting bit
   stream on standard output.  Setting ssize to 1 removes fake
   (i.e. does no decoding) for debugging purposes."

   option "ssize" m "symbol size (range [2,16], default is 8)" int no
   option "bsize" n "block size (range [1,(2^m)-1], default is 255)" int no
   option "dsize" k "data symbols (range is [1,n], default is 223)" int no
   option "debug" d "print some more error messages" flag off
***** -= gengetopt end =- ************************************/

int
main (int argc, char *argv[])
{
  bool debug = FALSE;			/* printout */
  int mm = 8;				/* symbol size */
  int nn = 255;				/* block size */
  int kk = 223;				/* information symbols per block */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num != 0)
    {
      parseopts_print_help();
      fatal("no arguments required", NULL);
    }
  if (args_info.ssize_given)
    {
      mm = args_info.ssize_arg;
      if (mm < 1 || mm > 16)
  	fatal("symbol size is out of range", NULL);
    }
  if (args_info.bsize_given)
    {
      nn = args_info.bsize_arg;
      if (nn < 1 || nn > (1 << mm) - 1)
	fatal("block size is out of range", NULL);
    }
  if (args_info.dsize_given)
    {
      kk = args_info.dsize_arg;
      if (kk < 1 || kk > nn)
	fatal("data symbols are out of range", NULL);
    }
  if (args_info.debug_given)
    debug = TRUE;

  (void) init_stdio_bufs();


  /* Initialise */
  bool fast_decoder = FALSE;
  if (mm == 8 && nn == 255 && kk == 223)
    fast_decoder = TRUE;


  /* Begin transmission */
  if (mm == 1)				/* fake decoder */
    {
      byte x[nn];			/* data block */
      bzero(x, nn);
      while (nn == fread(x, 1, nn, stdin))
	if (kk != fwrite(x, 1, kk, stdout))
	  pfatal("error writing on output");
    }
  else if (fast_decoder)
    {
      byte x[nn];			/* data block */
      while (nn == fread(x, 1, nn, stdin))
	{
	  (void) decode_rs_8(x, NULL, 0);
	  if (kk != fwrite(x, 1, kk, stdout))
	    pfatal("error writing on output");
	}
    }
  else
    fatal("code not implemented", NULL);

  if (debug && feof(stdin))
    error("end of file reading from input", NULL);
  if (ferror(stdin))
    pfatal("error reading from input");


  return 0;
}


/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
