/*
 * A suite of programs for simulating channel codings
 * Copyright (C) 2002 Francesco Potort
 * See file COPYING for details
 *
 * Most of this file is code from the 3.0.1 viterbi package:
 * Copyright 1999 Phil Karn, KA9Q
 */

#include "codec.h"
#include "lib/conv.h"
#include "lib/vdecoder_opt.h"

const char *progname = PARSEOPTS_PACKAGE;

static int mergedist;			/* decoding memory depth */
static int pathmem;			/* better if power of 2 */
static int tracechunk;			/* decode this many bits at a time */


static void traceback (unsigned long paths[], unsigned int pi);
static void flush (unsigned long paths[], unsigned int pi, int bitcnt);

/** -= gengetopt optstart =- *******************
   --func-name=parseopts --unamed-opts
*** -= gengetopt datastart =- ******************
   package "vdecoder"

   version "1.21"

   purpose "Reads a sequence of binary offset byte values from standard input.
   0 is illegal, [1-127] is a -1, 128 is an erasure, [129-255] is a 1.
   Decode the stream using the Viterbi algorithm and write it on standard output
   as a bit stream in network byte order."

   option "code" c "use the given known code (default=7,1/2)" string no
   option "list" l "list the known codes and exit" flag off
   option "decmem" m "decoding memory depth (default=80)" int no
   option "debug" d "print some more error messages" flag off
*** -= gengetopt end =- ***********************/


int
main (int argc, char *argv[])
{
  bool debug = FALSE;			/* printout */
  int K = 7;				/* constraint length */
  int k = 1;				/* code dimension */
  int n = 2;				/* codeword length */
  unsigned int *taps = NULL;		/* the n polynomials */
  mergedist = 80;			/* decoding memory depth */

  struct gengetopt_args_info args_info;
  (void) parseopts(argc, argv, &args_info);

  if (args_info.inputs_num != 0)
    {
      parseopts_print_help();
      fatal("no arguments required", NULL);
    }
  if (args_info.code_given)
    {
      if (sscanf(args_info.code_arg, "%u,%u/%u", &K, &k, &n) != 3)
  	fatal("wrong format for code", NULL);
    }
  if (args_info.decmem_given)
    {
      mergedist = args_info.decmem_arg;
      if (mergedist < K || mergedist > 1024)
	fatal("strange value for decoding memory depth", NULL);
    }
  if (args_info.list_given)
    {
      list_ccodes(stdout);		/* do a listing of the known codes */
      exit(0);				/* and nothing more */
    }
  if (args_info.debug_given)
    debug = TRUE;

  (void) init_stdio_bufs();


  /* Initialisation */
  pathmem = 2;
  for (int shift = 0; (mergedist >> shift) > 0; shift++)
    pathmem <<= 1;
  tracechunk = (pathmem - mergedist) & ~7;
  for (struct ccode *cp = known_ccodes; cp->K > 0; cp++)
    if (cp->K == K && cp->k == k && cp->n == n)
      taps = cp->taps;
  if (taps == NULL)
    fatal("no known code for (%.9s)", args_info.code_arg);


  /* Begin reception */
  if (K == 7 && k == 1 && n == 2	/* (7,1/2) code */
      && taps[0] == 0x6d && taps[1] == 0x4f) /* the standard one */
    {
      /* This code copied from vitfilt27.c by Phil Karn, which see. */

      /* Initialize metric table (make this an option)
       * This table assumes a symbol of 0 is the
       * strongest possible '0', and a symbol
       * of 255 is the strongest possible '1'. A symbol
       * of 128 is an erasure
       */
      int mettab[2][256];
      for (int i = 0; i < 256; i++)
	{
	  mettab[0][i] = OFFSET - i;
	  mettab[1][255 - i] = OFFSET - i - 1;
	}

      long cmetric[64], nmetric[64];
      cmetric[0] = 0;
      for (int i = 1; i < 64; i++)
	cmetric[i] = -99999;

      /* Main loop -- read input symbols and run ACS butterflies,
       * periodically tracing back to produce decoded output data.
       * The loop is unrolled to process two bits per iteration.
       */
      unsigned long paths[2 * pathmem];
      int bitcnt = -mergedist;
      unsigned int pi = 0;
      for (;;)
	{
	  /* Renormalize metrics to prevent overflow */
	  const int renormalize = 100000;
	  if (cmetric[0] > (LONG_MAX - renormalize))
	    for (int i = 0; i < 64; i++)
	      cmetric[i] -= LONG_MAX;
	  else if (cmetric[0] < LONG_MIN + renormalize)
	    for (int i = 0; i < 64; i++)
	      cmetric[i] += LONG_MAX;

	  register unsigned long dec;
	  int mets[4];
	  unsigned char symbols[2];

	  /* Read input symbol pair and compute branch metrics */
	  symbols[0] = getchar ();
	  symbols[1] = getchar ();
	  if (feof(stdin) || ferror(stdin))
	    break;
	  if (symbols[0] == 0 || symbols[1] == 0)
	    fatal("input stream is not in binary-offset mode", NULL);

	  mets[0] = mettab[0][symbols[0]] + mettab[0][symbols[1]];
	  mets[1] = mettab[0][symbols[0]] + mettab[1][symbols[1]];
	  mets[3] = mettab[1][symbols[0]] + mettab[1][symbols[1]];
	  mets[2] = mettab[1][symbols[0]] + mettab[0][symbols[1]];

	  /* On even numbered bits, the butterflies read from cmetrics[]
	   * and write to nmetrics[]. On odd numbered bits, the reverse
	   * is done
	   */
	  dec = 0;
	  BFA_712 (0, 0);  BFA_712 (6, 0);  BFA_712 (8, 0);  BFA_712 (14, 0);
	  BFA_712 (2, 3);  BFA_712 (4, 3);  BFA_712 (10, 3); BFA_712 (12, 3);
	  BFA_712 (1, 1);  BFA_712 (7, 1);  BFA_712 (9, 1);  BFA_712 (15, 1);
	  BFA_712 (3, 2);  BFA_712 (5, 2);  BFA_712 (11, 2); BFA_712 (13, 2);
	  paths[2 * pi] = dec;
	  dec = 0;
	  BFA_712 (19, 0); BFA_712 (21, 0); BFA_712 (27, 0); BFA_712 (29, 0);
	  BFA_712 (17, 3); BFA_712 (23, 3); BFA_712 (25, 3); BFA_712 (31, 3);
	  BFA_712 (18, 1); BFA_712 (20, 1); BFA_712 (26, 1); BFA_712 (28, 1);
	  BFA_712 (16, 2); BFA_712 (22, 2); BFA_712 (24, 2); BFA_712 (30, 2);
	  paths[2 * pi + 1] = dec;
	  pi++;

	  /* Read input symbol pair and compute branch metrics */
	  symbols[0] = getchar ();
	  symbols[1] = getchar ();
	  if (feof(stdin) || ferror(stdin))
	    break;
	  if (symbols[0] == 0 || symbols[1] == 0)
	    fatal("input stream is not in binary-offset mode", NULL);

	  mets[0] = mettab[0][symbols[0]] + mettab[0][symbols[1]];
	  mets[1] = mettab[0][symbols[0]] + mettab[1][symbols[1]];
	  mets[3] = mettab[1][symbols[0]] + mettab[1][symbols[1]];
	  mets[2] = mettab[1][symbols[0]] + mettab[0][symbols[1]];

	  dec = 0;
	  BFB_712 (0, 0); BFB_712 (6, 0); BFB_712 (8, 0); BFB_712 (14, 0);
	  BFB_712 (2, 3); BFB_712 (4, 3); BFB_712 (10, 3); BFB_712 (12, 3);
	  BFB_712 (1, 1); BFB_712 (7, 1); BFB_712 (9, 1); BFB_712 (15, 1);
	  BFB_712 (3, 2); BFB_712 (5, 2); BFB_712 (11, 2); BFB_712 (13, 2);
	  paths[2 * pi] = dec;
	  dec = 0;
	  BFB_712 (19, 0); BFB_712 (21, 0); BFB_712 (27, 0); BFB_712 (29, 0);
	  BFB_712 (17, 3); BFB_712 (23, 3); BFB_712 (25, 3); BFB_712 (31, 3);
	  BFB_712 (18, 1); BFB_712 (20, 1); BFB_712 (26, 1); BFB_712 (28, 1);
	  BFB_712 (16, 2); BFB_712 (22, 2); BFB_712 (24, 2); BFB_712 (30, 2);
	  paths[2 * pi + 1] = dec;
	  pi = (pi + 1) % pathmem;
	  if ((bitcnt += 2) == tracechunk)
	    {
	      traceback(paths, pi);
	      bitcnt = 0;
	    }
	}

      flush(paths, pi, bitcnt);
    } /* the standard (7,1/2) code */
  else
    fatal("code not implemented", NULL);

  if (debug && feof(stdin))
    error("end of file reading from input", NULL);
  if (ferror(stdin))
    pfatal("error reading from input");

  return 0;
}


/* Periodic traceback to produce decoded data */
static void
traceback (unsigned long paths[], unsigned int pi)
{
  int beststate;
  unsigned char data[tracechunk / 8];

  /* Start on an arbitrary path and trace it back until it's almost
   * certain we've merged onto the best path
   */
  beststate = 0;		/* arbitrary */
  pi = (pi - 1) % pathmem;	/* Undo last increment of pi */
  for (int i = 0; i < mergedist - 6; i++)
    {
      if (paths[2 * pi + (beststate >> 5)] & (1 << (beststate & 31)))
	beststate |= 64;	/* 2^(K-1) */
      beststate >>= 1;
      pi = (pi - 1) % pathmem;
    }
  /* bestpath is now the encoder state on the best path, mergedist
   * bits back. We continue to chain back until we accumulate
   * tracechunk bits of decoded data
   */
  for (int j = sizeof (data) - 1; j >= 0; j--)
    {
      data[j] = 0;
      for (int i = 0; i < 8; i++)
	{
	  if (paths[2 * pi + (beststate >> 5)] & (1 << (beststate & 31)))
	    {
	      beststate |= 64;	/* 2^(K-1) */
	      data[j] |= 1 << i;
	    }
	  beststate >>= 1;
	  pi = (pi - 1) % pathmem;
	}
    }
  if (sizeof(data) != fwrite (data, 1, sizeof(data), stdout))
    pfatal("error writing on output");
}

/* Final traceback at end of trellis, assuming sender tailed to zero after an
 * integral number of data bytes. Trailing bits are dropped.
 */
static void
flush (unsigned long paths[], unsigned int pi, int bitcnt)
{
  int beststate, n, off;
  unsigned char data[pathmem / 8];

  beststate = 0;		/* Assume encoder tailing to 0 state */
  n = (mergedist - 6 + bitcnt) / 8;
  off = (mergedist - 6 + bitcnt) % 8;	/* ignored partial byte */
  pi = (pi - off - 1) % pathmem;
  for (int j = n - 1; j >= 0; j--)
    {
      data[j] = 0;
      for (int i = 0; i < 8; i++)
	{
	  if (paths[2 * pi + (beststate >> 5)] & (1 << (beststate & 31)))
	    {
	      beststate |= 64;	/* 2^(K-1) */
	      data[j] |= 1 << i;
	    }
	  beststate >>= 1;
	  pi = (pi - 1) % pathmem;
	}
    }
  if (n != fwrite (data, 1, n, stdout))
    pfatal("error writing on output");
}


/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE" "\\sw+_t" "bigcnt" "uint32" "byte" "bool")
 * End:
 */
